"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

import logging
from abc import ABCMeta, abstractmethod

from .PayloadPriority import DEFAULT_PRIORITY
from .VortexServer import VortexServer

logger = logging.getLogger(name=__name__)


class VortexConnectionABC(metaclass=ABCMeta):
    def __init__(
        self,
        logger,
        vortexServer: VortexServer,
        remoteVortexUuid: str,
        remoteVortexName: str,
        httpSessionUuid=None,
    ) -> None:
        self._vortexServer = vortexServer
        self._logger = logger
        self._remoteVortexUuid = remoteVortexUuid
        self._remoteVortexName = remoteVortexName
        self._closed = False
        self._httpSessionUuid = httpSessionUuid

    @property
    def httpSessionUuid(self):
        return self._httpSessionUuid

    @property
    def remoteVortexUuid(self):
        """Vortex UUID

        The vortex UUID of the remote vortex which this connection represents.

        """
        return self._remoteVortexUuid

    @property
    def remoteVortexName(self):
        """Vortex Name

        The name of the remote vortex which this connection represents.

        """
        return self._remoteVortexName

    @abstractmethod
    def write(self, payloadVortexStr: bytes, priority: int = DEFAULT_PRIORITY):
        """Write

        EG

        > assert not self._closed
        > self._request.write(payloadVortexStr)
        > self._request.write('.')

        """

    def close(self):
        self._vortexServer.connectionClosed(self)
        self._closed = True



"""
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  Synerty Pty Ltd
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
 *
"""

from vortex.Tuple import TupleField, Tuple, addTupleType


@addTupleType
class PerformTestActionTuple(Tuple):
    """Perform Test Action Tuple

    This tuple is used for testing the action code.

    """

    __tupleType__ = "synerty.vortex.PerformTestActionTuple"
    actionDataInt = TupleField(typingType=int)
    actionDataUnicode = TupleField(typingType=str)
    failProcessing = TupleField(typingType=bool)



import logging
import typing
import weakref
from abc import abstractmethod, ABCMeta
from collections import defaultdict
from datetime import datetime, timedelta
from typing import Optional, Dict, Set

import pytz
from rx.subjects import Subject
from twisted.internet import task

from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.TupleSelector import TupleSelector
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)


class _CachedSubscribedData:
    """Cached Subscribed Data

    The client will now cache
    """

    TEARDOWN_WAIT = 120  # 2 minutes in seconds

    __memoryLoggingRefs = None
    __memoryLoggingEnabled = False

    @classmethod
    def setupMemoryLogging(cls) -> None:
        cls.__memoryLoggingRefs = []
        cls.__memoryLoggingEnabled = True

    @classmethod
    def memoryLoggingDump(
        cls, top=10, over=100
    ) -> typing.List[typing.Tuple[str, int, int]]:
        if not cls.__memoryLoggingRefs:
            return []

        # Filter out expired items
        cls.__memoryLoggingRefs = list(
            filter(lambda o: o(), cls.__memoryLoggingRefs)
        )

        cachesByTupleType = defaultdict(list)
        for cacheRef in cls.__memoryLoggingRefs:
            cache = cacheRef()
            if not cache:
                continue
            cachesByTupleType[cache.tupleSelector.name].append(
                len(cache.encodedPayload)
            )

        results = []
        for tupleName, lengths in cachesByTupleType.items():
            results.append((tupleName, len(lengths), sum(lengths)))

        data = sorted(results, key=lambda x: x[2], reverse=True)

        return list(filter(lambda x: x[2] >= over, data))[:top]

    def __init__(
        self, tupleSelector: TupleSelector, cacheEnabled: bool = True
    ) -> None:
        if _CachedSubscribedData.__memoryLoggingEnabled:
            _CachedSubscribedData.__memoryLoggingRefs.append(weakref.ref(self))

        self.tupleSelector: TupleSelector = tupleSelector
        self.vortexUuids: Set[str] = set()
        self.tearDownDate: Optional[datetime] = None
        self.encodedPayload: bytes = None

        #: Is the cache enabled
        # HINT: Once it's turned off for a tupleSelector, it remains off.
        # Disabling it can also be done from VortexJS
        self.cacheEnabled = cacheEnabled

        #: Last Server Payload Date
        # If the server has responded with a payload, this is the date in the payload
        # @type {Date | null}
        self.lastServerPayloadDate: Optional[datetime] = None

        self.subject = Subject()

    def markForTearDown(self) -> None:
        if self.tearDownDate is not None:
            return
        self.tearDownDate = datetime.now(pytz.utc) + timedelta(
            seconds=self.TEARDOWN_WAIT
        )

    def resetTearDown(self) -> None:
        self.tearDownDate = None

    def isReadyForTearDown(self) -> bool:
        return (
            self.tearDownDate is not None
            and self.tearDownDate <= datetime.now(pytz.utc)
        )


class TupleDataObservableCache(metaclass=ABCMeta):
    __CHECK_PERIOD = 30  # seconds

    def __init__(self) -> None:
        self.__cache: Dict[str, _CachedSubscribedData] = {}

    def shutdown(self):
        if not self.__pollLoopingCall:
            raise Exception("This has already been shutdown")

        self.__cache = {}
        self.__pollLoopingCall.stop()
        self.__pollLoopingCall = None

        for cache in self.__cache.values():
            cache.subject.dispose()

        self.__cache = {}

    def start(self):

        self.__pollLoopingCall = task.LoopingCall(self.__cacheCheck)

        d = self.__pollLoopingCall.start(self.__CHECK_PERIOD, now=False)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def _tupleSelectors(self) -> typing.List[TupleSelector]:
        tupleSelectors = []
        for key in self.__cache:
            tupleSelectors.append(TupleSelector.fromJsonStr(key))
        return tupleSelectors

    ## ----- Implement local observable

    def __cacheCheck(self):
        currentVortexUuids = set(VortexFactory.getRemoteVortexUuids())

        for ts, cache in list(self.__cache.items()):
            cache.vortexUuids = cache.vortexUuids & currentVortexUuids

            if cache.vortexUuids or cache.subject.observers:
                cache.resetTearDown()

            elif cache.isReadyForTearDown():
                logger.debug("Cleaning cache for %s", cache.tupleSelector)
                self._sendUnsubscribeToServer(cache.tupleSelector)
                del self.__cache[ts]

            else:
                cache.markForTearDown()

    def _getCache(
        self, tupleSelector: TupleSelector
    ) -> Optional[_CachedSubscribedData]:
        return self.__cache.get(tupleSelector.toJsonStr())

    def _hasTupleSelector(self, tupleSelectorAny: TupleSelector) -> bool:
        return tupleSelectorAny.toJsonStr() in self.__cache

    def _makeCache(self, tupleSelector: TupleSelector) -> _CachedSubscribedData:
        tsStr = tupleSelector.toJsonStr()
        if tsStr in self.__cache:
            return self.__cache[tsStr]

        cache = _CachedSubscribedData(tupleSelector)
        self.__cache[tsStr] = cache
        return cache

    def _updateCache(
        self, payloadEnvelope: PayloadEnvelope
    ) -> typing.Tuple[_CachedSubscribedData, bool]:
        """Update Cache

        Update the cache if it requires it

        :returns a tuple of (cache, requiredUpdate)

        """
        tupleSelector = payloadEnvelope.filt["tupleSelector"]

        cache = self._getCache(tupleSelector)
        if not cache:
            cache = self._makeCache(tupleSelector)

        if cache.lastServerPayloadDate is not None:
            if payloadEnvelope.date < cache.lastServerPayloadDate:
                return cache, False

        cache.lastServerPayloadDate = payloadEnvelope.date
        cache.encodedPayload = payloadEnvelope.encodedPayload

        return cache, True

    @abstractmethod
    def _sendUnsubscribeToServer(self, tupleSelector: TupleSelector):
        pass



import logging
from datetime import datetime

import pytz
from twisted.internet import task

from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import (
    TuplesProviderABC,
    TupleDataObservableHandler,
)
from vortex.test.TestTuple import TestTuple
from vortex.test.TupleDataForTest import makeTestTupleData

logger = logging.getLogger(__name__)
testTuples1Selector = TupleSelector("testTuples1", {"count": 4})

testTuples2Selector = TupleSelector("testTuples2", {"count": 7})


class TestTupleProvider(TuplesProviderABC):
    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict, tupleSelector: TupleSelector) -> bytes:
        count = tupleSelector.selector["count"]
        tuples = makeTestTupleData(count)

        for t in tuples:
            t.aDate = datetime.now(pytz.utc)
            t.aDict = tupleSelector.selector
            t.aString = tupleSelector.name

        return (
            Payload(filt=filt, tuples=tuples)
            .makePayloadEnvelope()
            .toVortexMsg()
        )


class NotifyTestTimer:
    @classmethod
    def __notify(cls):
        observableHandler.notifyOfTupleUpdate(testTuples1Selector)
        observableHandler.notifyOfTupleUpdate(testTuples2Selector)

    @classmethod
    def startTupleUpdateNotifyer(cls):
        cls.__loopingCall = task.LoopingCall(cls.__notify)
        d = cls.__loopingCall.start(2)
        d.addErrback(lambda f: logger.exception(f.value))
        d.addCallback(
            lambda _: logger.debug("Observable tuple updates started")
        )


observableHandler = TupleDataObservableHandler("vortexTestObservable")
observableHandler.addTupleProvider(TestTuple.tupleName(), TestTupleProvider())
observableHandler.addTupleProvider("testTuples1", TestTupleProvider())
observableHandler.addTupleProvider("testTuples2", TestTupleProvider())



from abc import ABCMeta, abstractmethod
from typing import Optional, Union, Callable, List

from twisted.internet.defer import Deferred

from vortex.PayloadEnvelope import VortexMsgList

SendVortexMsgResponseCallable = Callable[
    [Union[VortexMsgList, bytes], int], Deferred
]


class VortexInfo:
    """Vortex Info

    This class stores details about the end of a vortex.
    """

    def __init__(self, name, uuid):
        self.name, self.uuid = name, uuid

    def __repr__(self):
        return "VortexInfo, Name=%s, UUID=%s" % (self.name, self.uuid)


class VortexABC(metaclass=ABCMeta):
    @property
    @abstractmethod
    def localVortexInfo(self) -> VortexInfo:
        pass

    @property
    @abstractmethod
    def remoteVortexInfo(self) -> List[VortexInfo]:
        pass

    @property
    @abstractmethod
    def requiresBase64Encoding(self) -> bool:
        pass

    @abstractmethod
    def sendVortexMsg(
        self,
        vortexMsgs: Union[VortexMsgList, bytes],
        vortexUuid: Optional[str] = None,
    ):
        """Send Vortex Msg

        Sends the vortex message to any conencted clients with vortexUuid.
        Or broadcast it to all connected vortex clients if it's None

        :param vortexMsgs: The vortex message(s) to send
        :param vortexUuid: The vortexUuid of the client to send to.
        """
